﻿using ImageMagick;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace Gifitti.Models
{
    /// <summary>
    /// Software model of a GIF, Encapsulated with bassic opperations on GIF images.
    /// </summary>
    public class GifModel
    {
        public Image gifImage { get; set; }
        public Image originalGif { get; private set; }
        private FrameDimension dimension;
        public int numberOfFrames { get; private set; }
        private Image[] frames;
        private int step = 1;
        private int currentFrame = -1;
        private bool reverse;
        public int startFrame = 0;
        public int endFrame = 0;

        /// <summary>
        /// Load a GIF into the software model from the system path "path"
        /// </summary>
        /// <param name="path">Location of the GIF being loaded.</param>
        public GifModel(String path)
        {
            //Save the original Gif
            originalGif = Image.FromFile(path);
            this.gifImage = originalGif.Clone() as Image;
            dimension = new FrameDimension(originalGif.FrameDimensionsList[0]);
            frameConstruction(originalGif);
        }

        /// <summary>
        /// Breaks GIF images into their frames.
        /// </summary>
        /// <param name="img">GIF image being decomposed.</param>
        private void frameConstruction(Image img)
        {
            numberOfFrames = img.GetFrameCount(FrameDimension.Time);
            endFrame = numberOfFrames;
            frames = new Image[numberOfFrames];

            for(int i = 0; i < numberOfFrames; i++)
            {
                originalGif.SelectActiveFrame(FrameDimension.Time, i);
                frames[i] = ((Image)originalGif.Clone());
            }
        }

        /// <summary>
        /// Changes the loaded GIF to be what was originally loaded by the User.
        /// If a new file is opened the current original GIF is overloaded.
        /// </summary>
        public void resetToOriginalGif()
        {
            gifImage = originalGif.Clone() as Image;
        }

        /// <summary>
        /// Takes the full path of a GIF save location, 
        /// extension included
        /// </summary>
        /// <param name="path">String representation of path to save the gif to. 
        ///     Name and extension of gif mandatory.
        /// </param>
        /// <exception cref="DirectoryNotFoundException">
        /// Thrown if the path is not valid for the local system,
        /// OR the file has an extension other than [G|g][I|i][F|f]
        /// </exception>
        public void saveGif(string path)
        {
            //Error Handel on path
            //RE for path with extension
            Regex extensionTester = new Regex(@"^.*\.[g|G][i|I][f|F]$");
            //RE for path without extension
            Regex savePathTester = new Regex(@"^((?:(?:[^\/\\]+)*[\/\\]*)*)[\/\\]([^\/\\]*)$");
            //if (!savePathTester.IsMatch(path) || !Directory.Exists(savePathTester.Match(path).Groups[1].ToString()))
            //    throw new DirectoryNotFoundException();
            //Handel non gif extensions
            if (!extensionTester.IsMatch(path))
                path += ".gif"; 
            //TODO take in image setting components as well as frames being viewed etc.
            //Build GIF
            using (MagickImageCollection collection = new MagickImageCollection())
            {
                for (int i = this.startFrame; i < Math.Min(this.endFrame, frames.Length); i++)
                    collection.Add(new MagickImage(frames[i] as Bitmap));
                
                // Maximize Colour saturation
                QuantizeSettings settings = new QuantizeSettings();
                settings.Colors = 256;
                collection.Quantize(settings);

                // Optimized the GIF being exported
                collection.Optimize();
                // Save GIF to path
                collection.Write(path);
            }
        }



        /// <summary>
        /// Fetches a sub gif over frames start -> stop
        /// </summary>
        /// <param name="start">Initial Frame.</param>
        /// <param name="stop">End Frame.</param>
        public void returnSubGif(int start, int stop)
        {
            using (MagickImageCollection collection = new MagickImageCollection())
            {
                // Add first image and set the animation delay to 100ms
                for (int i = start; i < stop; i++)
                {
                    collection.Add(new ImageMagick.MagickImage(frames[i] as Bitmap));
                    collection[i - start].AnimationDelay = 1;
                }
                // Optionally reduce colors
                QuantizeSettings settings = new QuantizeSettings();
                settings.Colors = 256;
                collection.Quantize(settings);

                // Optionally optimize the images (images should have the same size).
                collection.Optimize();

                // Save gif
                collection.Write("temp.gif");
            }
        }


        /// <summary>
        /// Whether the gif should play backwards when it reaches the end
        /// </summary>
        public bool ReverseAtEnd
        {
            get { return reverse; }
            set { reverse = value; }
        }

        /// <summary>
        /// Fetches the GIFs next viewable frame
        /// </summary>
        /// <returns>Next frame to be returned to the control.</returns>
        public Image GetNextFrame()
        {

            currentFrame += step;

            //if the animation reaches a boundary...
            if (currentFrame >= numberOfFrames || currentFrame < 1 || currentFrame >= endFrame)
            {
                if (reverse)
                {
                    step *= -1;
                    //...reverse the count
                    //apply it
                    currentFrame += step;
                }
                else
                {
                    currentFrame = startFrame;
                    //...or start over
                }
            }
            return GetFrame(currentFrame);
        }

        /// <summary>
        /// Gets the frame index.
        /// </summary>
        /// <param name="index">Frame number to fetch</param>
        /// <returns>Frame number index in the model</returns>
        public Image GetFrame(int index)
        {
            originalGif.SelectActiveFrame(dimension, index);
            //find the frame
            return (Image)originalGif.Clone();
            //return a copy of it
        }

        public void SetFrameFromImage(Image frame, int index)
        {
            frames[index] = frame;
        }
    }
}
